import collections
from collections.abc import Sequence
from typing import NamedTuple

import supriya.osc
from supriya import HeaderFormat, SampleFormat, utils
from supriya.enums import RequestId

from .bases import Request, RequestBundle, Response


class BufferAllocateRequest(Request):
    """
    A /b_alloc request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferAllocateRequest(
        ...     buffer_id=23,
        ...     frame_count=512,
        ...     channel_count=2,
        ... )
        >>> request
        BufferAllocateRequest(
            buffer_id=23,
            channel_count=2,
            frame_count=512,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_alloc', 23, 512, 2)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_ALLOCATE

    ### INITIALIZER ###

    def __init__(
        self, buffer_id=None, frame_count=None, channel_count=None, callback=None
    ):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        self._frame_count = frame_count
        if channel_count is not None:
            channel_count = int(channel_count)
            assert 0 < channel_count
        self._channel_count = channel_count
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        frame_count = int(self.frame_count)
        channel_count = int(self.channel_count)
        contents = [request_id, buffer_id, frame_count, channel_count]
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def channel_count(self):
        return self._channel_count

    @property
    def callback(self):
        return self._callback

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def response_patterns(self):
        return ["/done", "/b_alloc", self.buffer_id], None


class BufferAllocateReadRequest(BufferAllocateRequest):
    """
    A /b_allocRead request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferAllocateReadRequest(
        ...     buffer_id=23,
        ...     file_path="pulse_44100sr_16bit_octo.wav",
        ... )
        >>> print(request)
        BufferAllocateReadRequest(
            buffer_id=23,
            file_path='pulse_44100sr_16bit_octo.wav',
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_allocRead', 23, '...pulse_44100sr_16bit_octo.wav', 0, -1)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_ALLOCATE_READ

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        callback=None,
        file_path=None,
        frame_count=None,
        starting_frame=None,
    ):
        import supriya.nonrealtime

        BufferAllocateRequest.__init__(
            self, buffer_id=buffer_id, frame_count=frame_count, callback=callback
        )
        if not supriya.nonrealtime.Session.is_session_like(file_path):
            file_path = str(file_path)
        self._file_path = file_path
        if starting_frame is not None:
            starting_frame = int(starting_frame)
            assert 0 <= starting_frame
        self._starting_frame = starting_frame

    ### PRIVATE METHODS ###

    def _get_osc_message_contents(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        frame_count = self.frame_count
        if frame_count is None:
            frame_count = -1
        starting_frame = self.starting_frame
        if starting_frame is None:
            starting_frame = 0
        contents = [request_id, buffer_id, self.file_path, starting_frame, frame_count]
        return contents

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        contents = self._get_osc_message_contents()
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def callback(self):
        return self._callback

    @property
    def file_path(self):
        return self._file_path

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def response_patterns(self):
        return ["/done", "/b_allocRead", self.buffer_id], None

    @property
    def starting_frame(self):
        return self._starting_frame


class BufferAllocateReadChannelRequest(BufferAllocateReadRequest):
    """
    A /b_allocRead request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferAllocateReadChannelRequest(
        ...     buffer_id=23,
        ...     channel_indices=(3, 4),
        ...     file_path="pulse_44100sr_16bit_octo.wav",
        ... )
        >>> print(request)
        BufferAllocateReadChannelRequest(
            buffer_id=23,
            channel_indices=(3, 4),
            file_path='pulse_44100sr_16bit_octo.wav',
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_allocReadChannel', 23, '...pulse_44100sr_16bit_octo.wav', 0, -1, 3, 4)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_ALLOCATE_READ_CHANNEL

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        channel_indices=None,
        callback=None,
        file_path=None,
        frame_count=None,
        starting_frame=None,
    ):
        BufferAllocateReadRequest.__init__(
            self,
            buffer_id=buffer_id,
            callback=callback,
            file_path=file_path,
            frame_count=frame_count,
            starting_frame=starting_frame,
        )
        if channel_indices is None:
            channel_indices = -1
        if not isinstance(channel_indices, Sequence):
            channel_indices = (channel_indices,)
        channel_indices = tuple(int(_) for _ in channel_indices)
        if channel_indices != (-1,):
            assert all(0 <= _ for _ in channel_indices)
        self._channel_indices = channel_indices

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        contents = self._get_osc_message_contents()
        contents.extend(self.channel_indices)
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def channel_indices(self):
        return self._channel_indices

    @property
    def response_patterns(self):
        return ["/done", "/b_allocReadChannel", self.buffer_id], None


class BufferCloseRequest(Request):
    """
    A /b_close request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferCloseRequest(
        ...     buffer_id=23,
        ... )
        >>> request
        BufferCloseRequest(
            buffer_id=23,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_close', 23)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_CLOSE

    ### INITIALIZER ###

    def __init__(self, buffer_id=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        message = supriya.osc.OscMessage(request_id, buffer_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def response_patterns(self):
        return ["/done", "/b_close", self.buffer_id], None


class BufferCopyRequest(Request):
    """
    A `/b_gen copy` request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferCopyRequest(
        ...     source_buffer_id=23,
        ...     target_buffer_id=666,
        ... )
        >>> print(request)
        BufferCopyRequest(
            source_buffer_id=23,
            target_buffer_id=666,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_gen', 666, 'copy', 0, 23, 0, -1)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_GENERATE

    ### INITIALIZER ###

    def __init__(
        self,
        frame_count=None,
        source_buffer_id=None,
        source_starting_frame=None,
        target_buffer_id=None,
        target_starting_frame=None,
    ):
        Request.__init__(self)
        self._source_buffer_id = int(source_buffer_id)
        self._target_buffer_id = int(target_buffer_id)
        if frame_count is not None:
            frame_count = int(frame_count)
            assert -1 <= frame_count
        self._frame_count = frame_count
        if source_starting_frame is not None:
            source_starting_frame = int(source_starting_frame)
            assert 0 <= source_starting_frame
        self._source_starting_frame = source_starting_frame
        if target_starting_frame is not None:
            target_starting_frame = int(target_starting_frame)
            assert 0 <= target_starting_frame
        self._target_starting_frame = target_starting_frame

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        frame_count = self.frame_count
        if frame_count is None:
            frame_count = -1
        source_starting_frame = self.source_starting_frame
        if source_starting_frame is None:
            source_starting_frame = 0
        target_starting_frame = self.target_starting_frame
        if target_starting_frame is None:
            target_starting_frame = 0
        contents = [
            request_id,
            self.target_buffer_id,
            "copy",
            target_starting_frame,
            self.source_buffer_id,
            source_starting_frame,
            frame_count,
        ]
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def response_patterns(self):
        return ["/done", "/b_gen", self.target_buffer_id], None

    @property
    def source_buffer_id(self):
        return self._source_buffer_id

    @property
    def source_starting_frame(self):
        return self._source_starting_frame

    @property
    def target_buffer_id(self):
        return self._target_buffer_id

    @property
    def target_starting_frame(self):
        return self._target_starting_frame


class BufferFillRequest(Request):
    """
    A /b_fill request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferFillRequest(
        ...     buffer_id=23,
        ...     index_count_value_triples=(
        ...         (0, 8, 0.1),
        ...         (11, 4, 0.2),
        ...     ),
        ... )
        >>> request
        BufferFillRequest(
            buffer_id=23,
            index_count_value_triples=(
                (0, 8, 0.1),
                (11, 4, 0.2),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_fill', 23, 0, 8, 0.1, 11, 4, 0.2)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_FILL

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, index_count_value_triples=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        triples = []
        for index, count, value in index_count_value_triples:
            triple = (int(index), int(count), float(value))
            triples.append(triple)
        triples = tuple(triples)
        self._index_count_value_triples = triples

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        for index, count, value in self.index_count_value_triples:
            contents.append(int(index))
            contents.append(int(count))
            contents.append(float(value))
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def index_count_value_triples(self):
        return self._index_count_value_triples


class BufferFreeRequest(Request):
    """
    A /b_free request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferFreeRequest(
        ...     buffer_id=23,
        ... )
        >>> request
        BufferFreeRequest(
            buffer_id=23,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_free', 23)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_FREE

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, callback=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def callback(self):
        return self._callback


class BufferGenerateRequest(Request):
    """
    A /b_gen request.

    This requests models the 'cheby', 'sine1', 'sine2' and 'sine3' /b_gen
    commands.

    Use BufferCopyRequest for `/b_gen copy` and BufferNormalizeRequest for
    `/b_gen normalize` and `/b_gen wnormalize`.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferGenerateRequest(
        ...     amplitudes=(1.0, 0.5, 0.25),
        ...     as_wavetable=True,
        ...     buffer_id=23,
        ...     command_name="sine3",
        ...     frequencies=(1, 2, 3),
        ...     phases=(0, 0.5, 0),
        ...     should_clear_first=True,
        ...     should_normalize=True,
        ... )
        >>> print(request)
        BufferGenerateRequest(
            amplitudes=(1.0, 0.5, 0.25),
            as_wavetable=True,
            buffer_id=23,
            command_name='sine3',
            frequencies=(1.0, 2.0, 3.0),
            phases=(0.0, 0.5, 0.0),
            should_clear_first=True,
            should_normalize=True,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_gen', 23, 'sine3', 7, 1.0, 1.0, 0.0, 2.0, 0.5, 0.5, 3.0, 0.25, 0.0)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_GENERATE

    ### INITIALIZER ###

    def __init__(
        self,
        amplitudes=None,
        as_wavetable=None,
        buffer_id=None,
        command_name=None,
        frequencies=None,
        phases=None,
        should_clear_first=None,
        should_normalize=None,
    ):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        assert command_name in ("cheby", "sine1", "sine2", "sine3")
        self._command_name = command_name
        if as_wavetable is not None:
            as_wavetable = bool(as_wavetable)
        self._as_wavetable = as_wavetable
        if should_clear_first is not None:
            should_clear_first = bool(should_clear_first)
        self._should_clear_first = should_clear_first
        if should_normalize is not None:
            should_normalize = bool(should_normalize)
        self._should_normalize = should_normalize
        self._frequencies = None
        self._phases = None
        if command_name in ("cheby", "sine1"):
            if not isinstance(amplitudes, Sequence):
                amplitudes = (amplitudes,)
            amplitudes = tuple(float(_) for _ in amplitudes)
            assert len(amplitudes)
            self._amplitudes = amplitudes
        if command_name == "sine2":
            amplitudes = tuple(float(_) for _ in amplitudes)
            frequencies = tuple(float(_) for _ in frequencies)
            assert 0 < len(amplitudes)
            assert len(amplitudes) == len(frequencies)
            self._amplitudes = amplitudes
            self._frequencies = frequencies
        if command_name == "sine3":
            amplitudes = tuple(float(_) for _ in amplitudes)
            frequencies = tuple(float(_) for _ in frequencies)
            phases = tuple(float(_) for _ in phases)
            assert 0 < len(amplitudes)
            assert len(amplitudes) == len(frequencies) == len(phases)
            self._amplitudes = amplitudes
            self._frequencies = frequencies
            self._phases = phases

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id, self.command_name, self.flags]
        if self.command_name in ("cheby", "sine1"):
            coefficients = self.amplitudes
        elif self.command_name == "sine2":
            coefficients = zip(self.frequencies, self.amplitudes)
            coefficients = tuple(coefficients)
        elif self.command_name == "sine3":
            coefficients = zip(self.frequencies, self.amplitudes, self.phases)
            coefficients = tuple(coefficients)
        coefficients = utils.flatten_iterable(coefficients)
        contents.extend(coefficients)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC METHODS ###

    @classmethod
    def chebyshev(
        cls,
        amplitudes=None,
        as_wavetable=False,
        buffer_id=None,
        should_normalize=True,
        should_clear_first=True,
    ):
        command_name = "cheby"
        request = cls(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=buffer_id,
            command_name=command_name,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        return request

    @classmethod
    def sine1(
        cls,
        amplitudes=None,
        as_wavetable=False,
        buffer_id=None,
        should_normalize=True,
        should_clear_first=True,
    ):
        command_name = "sine1"
        request = cls(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=buffer_id,
            command_name=command_name,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        return request

    @classmethod
    def sine2(
        cls,
        amplitudes=None,
        as_wavetable=False,
        buffer_id=None,
        frequencies=None,
        should_normalize=True,
        should_clear_first=True,
    ):
        command_name = "sine2"
        request = cls(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=buffer_id,
            command_name=command_name,
            frequencies=frequencies,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        return request

    @classmethod
    def sine3(
        cls,
        amplitudes=None,
        as_wavetable=False,
        buffer_id=None,
        frequencies=None,
        phases=None,
        should_normalize=True,
        should_clear_first=True,
    ):
        command_name = "sine3"
        request = cls(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=buffer_id,
            command_name=command_name,
            frequencies=frequencies,
            phases=phases,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        return request

    ### PUBLIC PROPERTIES ###

    @property
    def amplitudes(self):
        return self._amplitudes

    @property
    def as_wavetable(self):
        return self._as_wavetable

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def command_name(self):
        return self._command_name

    @property
    def flags(self):
        flags = sum(
            (
                1 * int(bool(self.should_normalize)),
                2 * int(bool(self.as_wavetable)),
                4 * int(bool(self.should_clear_first)),
            )
        )
        return flags

    @property
    def frequencies(self):
        return self._frequencies

    @property
    def phases(self):
        return self._phases

    @property
    def response_patterns(self):
        return ["/done", "/b_gen", self.buffer_id], None

    @property
    def should_clear_first(self):
        return self._should_clear_first

    @property
    def should_normalize(self):
        return self._should_normalize


class BufferGetContiguousRequest(Request):
    """
    A /b_getn request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferGetContiguousRequest(
        ...     buffer_id=23,
        ...     index_count_pairs=[(0, 3), (8, 11)],
        ... )
        >>> request
        BufferGetContiguousRequest(
            buffer_id=23,
            index_count_pairs=(
                (0, 3),
                (8, 11),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_getn', 23, 0, 3, 8, 11)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_GET_CONTIGUOUS

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, index_count_pairs=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        self._index_count_pairs = tuple(
            (int(index), int(count)) for index, count in index_count_pairs
        )

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.index_count_pairs:
            for index, count in self.index_count_pairs:
                contents.append(index)
                contents.append(count)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def index_count_pairs(self):
        return self._index_count_pairs

    @property
    def response_patterns(self):
        return ["/b_setn", self.buffer_id], ["/fail", "/b_getn"]


class BufferGetRequest(Request):
    """
    A /b_get request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferGetRequest(
        ...     buffer_id=23,
        ...     indices=(0, 4, 8, 16),
        ... )
        >>> request
        BufferGetRequest(
            buffer_id=23,
            indices=(0, 4, 8, 16),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_get', 23, 0, 4, 8, 16)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_GET

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, indices=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        self._indices = tuple(int(index) for index in indices)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.indices:
            for index in self.indices:
                contents.append(index)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def indices(self):
        return self._indices

    @property
    def response_patterns(self):
        return ["/b_set", self.buffer_id], ["/fail", "/b_get"]


class BufferInfoResponse(Response):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        buffer_id: int
        frame_count: int
        channel_count: int
        sample_rate: int

    ### INITIALIZER ###

    def __init__(self, items=None):
        self._items = tuple(items or ())

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        """
        Create response(s) from OSC message.

        ::

            >>> message = supriya.osc.OscMessage("/b_info", 1100, 512, 1, 44100.0)
            >>> supriya.commands.BufferInfoResponse.from_osc_message(message)
            BufferInfoResponse(
                items=(
                    Item(buffer_id=1100, frame_count=512, channel_count=1, sample_rate=44100.0),
                ),
            )

        """
        # TODO: Return one single thing
        items = []
        for group in cls._group_items(osc_message.contents, 4):
            item = cls.Item(*group)
            items.append(item)
        return cls(items=items)

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self[0].buffer_id

    @property
    def items(self):
        return self._items


class BufferNormalizeRequest(Request):
    """
    A `/b_gen normalize` request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferNormalizeRequest(
        ...     buffer_id=23,
        ... )
        >>> print(request)
        BufferNormalizeRequest(
            buffer_id=23,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_gen', 23, 'normalize', 1.0)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_GENERATE

    ### INITIALIZER ###

    def __init__(self, as_wavetable=None, buffer_id=None, new_maximum=1.0):
        Request.__init__(self)
        if as_wavetable is not None:
            as_wavetable = bool(as_wavetable)
        self._as_wavetable = as_wavetable
        self._buffer_id = int(buffer_id)
        self._new_maximum = float(new_maximum)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        command_name = "normalize"
        if self.as_wavetable:
            command_name = "wnormalize"
        contents = [request_id, buffer_id, command_name, self.new_maximum]
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def as_wavetable(self):
        return self._as_wavetable

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def new_maximum(self):
        return self._new_maximum

    @property
    def response_patterns(self):
        return ["/done", "/b_gen", self.buffer_id], None


class BufferQueryRequest(Request):
    """
    A /b_query request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferQueryRequest(buffer_ids=(1, 23, 41))
        >>> request
        BufferQueryRequest(
            buffer_ids=(1, 23, 41),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_query', 1, 23, 41)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_QUERY

    ### INITIALIZER ###

    def __init__(self, buffer_ids=None):
        Request.__init__(self)
        if buffer_ids:
            buffer_ids = tuple(int(buffer_id) for buffer_id in buffer_ids)
        self._buffer_ids = buffer_ids

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        for buffer_id in self.buffer_ids:
            contents.append(buffer_id)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_ids(self):
        return self._buffer_ids

    @property
    def response_patterns(self):
        if 1 == len(self.buffer_ids):
            return ["/b_info", self.buffer_ids[0]], None
        return None, None


class BufferReadRequest(Request):
    """
    A /b_read request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferReadRequest(
        ...     buffer_id=23,
        ...     file_path="pulse_44100sr_16bit_octo.wav",
        ... )
        >>> print(request)
        BufferReadRequest(
            buffer_id=23,
            file_path='pulse_44100sr_16bit_octo.wav',
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_read', 23, '...pulse_44100sr_16bit_octo.wav', 0, -1, 0, 0)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_READ

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        callback=None,
        file_path=None,
        frame_count=None,
        leave_open=None,
        starting_frame_in_buffer=None,
        starting_frame_in_file=None,
    ):
        import supriya.nonrealtime

        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback
        if not supriya.nonrealtime.Session.is_session_like(file_path):
            file_path = str(file_path)
        self._file_path = file_path
        if frame_count is not None:
            frame_count = int(frame_count)
            assert -1 <= frame_count
        self._frame_count = frame_count
        if leave_open is not None:
            leave_open = bool(leave_open)
        self._leave_open = leave_open
        if starting_frame_in_buffer is not None:
            starting_frame_in_buffer = int(starting_frame_in_buffer)
            assert 0 <= starting_frame_in_buffer
        self._starting_frame_in_buffer = starting_frame_in_buffer
        if starting_frame_in_file is not None:
            starting_frame_in_file = int(starting_frame_in_file)
            assert 0 <= starting_frame_in_file
        self._starting_frame_in_file = starting_frame_in_file

    ### PRIVATE METHODS ###

    def _get_osc_message_contents(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        starting_frame_in_buffer = self.starting_frame_in_buffer
        if starting_frame_in_buffer is None:
            starting_frame_in_buffer = 0
        frame_count = self.frame_count
        if frame_count is None:
            frame_count = -1
        starting_frame_in_file = self.starting_frame_in_file
        if starting_frame_in_file is None:
            starting_frame_in_file = 0
        leave_open = int(bool(self.leave_open))
        contents = [
            request_id,
            buffer_id,
            self.file_path,
            starting_frame_in_file,
            frame_count,
            starting_frame_in_buffer,
            leave_open,
        ]
        return contents

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        contents = self._get_osc_message_contents()
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def callback(self):
        return self._callback

    @property
    def file_path(self):
        return self._file_path

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def leave_open(self):
        return self._leave_open

    @property
    def response_patterns(self):
        return ["/done", "/b_read", self.buffer_id], None

    @property
    def starting_frame_in_buffer(self):
        return self._starting_frame_in_buffer

    @property
    def starting_frame_in_file(self):
        return self._starting_frame_in_file


class BufferReadChannelRequest(BufferReadRequest):
    """
    A /b_readChannel request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferReadChannelRequest(
        ...     buffer_id=23,
        ...     channel_indices=(3, 4),
        ...     file_path="pulse_44100sr_16bit_octo.wav",
        ... )
        >>> print(request)
        BufferReadChannelRequest(
            buffer_id=23,
            channel_indices=(3, 4),
            file_path='pulse_44100sr_16bit_octo.wav',
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_readChannel', 23, '...pulse_44100sr_16bit_octo.wav', 0, -1, 0, 0, 3, 4)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_READ_CHANNEL

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        channel_indices=None,
        callback=None,
        file_path=None,
        frame_count=None,
        leave_open=None,
        starting_frame_in_buffer=None,
        starting_frame_in_file=None,
    ):
        BufferReadRequest.__init__(
            self,
            buffer_id=buffer_id,
            callback=callback,
            file_path=file_path,
            frame_count=frame_count,
            leave_open=leave_open,
            starting_frame_in_buffer=starting_frame_in_buffer,
            starting_frame_in_file=starting_frame_in_file,
        )
        if not isinstance(channel_indices, Sequence):
            channel_indices = (channel_indices,)
        channel_indices = tuple(channel_indices)
        assert all(0 <= _ for _ in channel_indices)
        self._channel_indices = channel_indices

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        contents = self._get_osc_message_contents()
        contents.extend(self.channel_indices)
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def channel_indices(self):
        return self._channel_indices

    @property
    def response_patterns(self):
        return ["/done", "/b_readChannel", self.buffer_id], None


class BufferSetContiguousRequest(Request):
    """
    A /b_setn request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferSetContiguousRequest(
        ...     buffer_id=23,
        ...     index_values_pairs=((0, (1, 2, 3)), (10, (17.1, 18.2))),
        ... )
        >>> request
        BufferSetContiguousRequest(
            buffer_id=23,
            index_values_pairs=(
                (0, (1.0, 2.0, 3.0)),
                (10, (17.1, 18.2)),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_setn', 23, 0, 3, 1.0, 2.0, 3.0, 10, 2, 17.1, 18.2)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_SET_CONTIGUOUS

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, index_values_pairs=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if index_values_pairs:
            pairs = []
            for index, values in index_values_pairs:
                index = int(index)
                values = tuple(float(value) for value in values)
                pair = (index, values)
                pairs.append(pair)
            pairs = tuple(pairs)
        self._index_values_pairs = pairs

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.index_values_pairs:
            for index, values in self.index_values_pairs:
                if not values:
                    continue
                contents.append(index)
                contents.append(len(values))
                for value in values:
                    contents.append(value)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def index_values_pairs(self):
        return self._index_values_pairs


class BufferSetContiguousResponse(Response, Sequence):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        sample_values: int
        starting_sample_index: int

    ### INITIALIZER ###

    def __init__(self, items=None, buffer_id=None):
        self._buffer_id = buffer_id
        self._items = items

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    ### PUBLIC METHODS ###

    def as_dict(self):
        result = collections.OrderedDict()
        for item in self:
            result[item.starting_sample_index] = item.sample_values
        return result

    @classmethod
    def from_osc_message(cls, osc_message):
        """
        Create response from OSC message.

        ::

            >>> message = supriya.osc.OscMessage(
            ...     "/b_setn", 1, 0, 8, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
            ... )
            >>> supriya.commands.BufferSetContiguousResponse.from_osc_message(message)
            BufferSetContiguousResponse(
                buffer_id=1,
                items=(
                    Item(sample_values=(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0), starting_sample_index=0),
                ),
            )

        """
        buffer_id, remainder = osc_message.contents[0], osc_message.contents[1:]
        items = []
        while remainder:
            starting_sample_index = remainder[0]
            sample_count = remainder[1]
            sample_values = tuple(remainder[2 : 2 + sample_count])
            item = cls.Item(
                starting_sample_index=starting_sample_index, sample_values=sample_values
            )
            items.append(item)
            remainder = remainder[2 + sample_count :]
        items = tuple(items)
        response = cls(buffer_id=buffer_id, items=items)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def items(self):
        return self._items


class BufferSetRequest(Request):
    """
    A /b_set request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferSetRequest(
        ...     buffer_id=23,
        ...     index_value_pairs=(
        ...         (0, 1.0),
        ...         (10, 13.2),
        ...         (17, 19.3),
        ...     ),
        ... )
        >>> request
        BufferSetRequest(
            buffer_id=23,
            index_value_pairs=(
                (0, 1.0),
                (10, 13.2),
                (17, 19.3),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_set', 23, 0, 1.0, 10, 13.2, 17, 19.3)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_SET

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, index_value_pairs=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if index_value_pairs:
            pairs = []
            for index, value in index_value_pairs:
                index = int(index)
                value = float(value)
                pair = (index, value)
                pairs.append(pair)
            pairs = tuple(pairs)
        self._index_value_pairs = index_value_pairs

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.index_value_pairs:
            for index, value in self.index_value_pairs:
                contents.append(index)
                contents.append(value)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def index_value_pairs(self):
        return self._index_value_pairs


class BufferSetResponse(Response, Sequence):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        sample_index: int
        sample_value: float

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, items=None):
        self._buffer_id = buffer_id
        self._items = items

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    ### PUBLIC METHODS ###

    def as_dict(self):
        result = collections.OrderedDict()
        for item in self:
            result[item.sample_index] = item.sample_value
        return result

    @classmethod
    def from_osc_message(cls, osc_message):
        buffer_id, remainder = osc_message.contents[0], osc_message.contents[1:]
        items = []
        for group in cls._group_items(remainder, 2):
            item = cls.Item(*group)
            items.append(item)
        items = tuple(items)
        response = cls(buffer_id=buffer_id, items=items)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def items(self):
        return self._items


class BufferWriteRequest(Request):
    """
    A /b_write request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferWriteRequest(
        ...     buffer_id=23,
        ...     file_path="test.aiff",
        ...     header_format=supriya.HeaderFormat.AIFF,
        ...     sample_format=supriya.SampleFormat.INT24,
        ... )
        >>> request
        BufferWriteRequest(
            buffer_id=23,
            file_path='test.aiff',
            frame_count=-1,
            header_format=HeaderFormat.AIFF,
            sample_format=SampleFormat.INT24,
            starting_frame=0,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_write', 23, 'test.aiff', 'aiff', 'int24', -1, 0, 0)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_WRITE

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        callback=None,
        file_path=None,
        frame_count=None,
        header_format="aiff",
        leave_open=False,
        sample_format="int24",
        starting_frame=None,
    ):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback
        self._file_path = str(file_path)
        if frame_count is None:
            frame_count = -1
        frame_count = int(frame_count)
        assert -1 <= frame_count
        self._frame_count = frame_count
        self._header_format = HeaderFormat.from_expr(header_format)
        self._leave_open = bool(leave_open)
        self._sample_format = SampleFormat.from_expr(sample_format)
        if starting_frame is None:
            starting_frame = 0
        starting_frame = int(starting_frame)
        assert 0 <= starting_frame
        self._starting_frame = starting_frame

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        header_format = self.header_format.name.lower()
        sample_format = self.sample_format.name.lower()
        leave_open = int(bool(self.leave_open))
        contents = [
            request_id,
            buffer_id,
            self.file_path,
            header_format,
            sample_format,
            self.frame_count,
            self.starting_frame,
            leave_open,
        ]
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def callback(self):
        return self._callback

    @property
    def file_path(self):
        return self._file_path

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def header_format(self):
        return self._header_format

    @property
    def leave_open(self):
        return self._leave_open

    @property
    def response_patterns(self):
        return ["/done", "/b_write", self.buffer_id], None

    @property
    def sample_format(self):
        return self._sample_format

    @property
    def starting_frame(self):
        return self._starting_frame


class BufferZeroRequest(Request):
    """
    A /b_zero request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.BufferZeroRequest(
        ...     buffer_id=23,
        ... )
        >>> request
        BufferZeroRequest(
            buffer_id=23,
        )

    ::

        >>> request.to_osc()
        OscMessage('/b_zero', 23)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.BUFFER_ZERO

    ### INITIALIZER ###

    def __init__(self, buffer_id=None, callback=None):
        Request.__init__(self)
        self._buffer_id = int(buffer_id)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        buffer_id = int(self.buffer_id)
        contents = [request_id, buffer_id]
        if self.callback:
            contents.append(self.callback.to_osc())
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        return self._buffer_id

    @property
    def callback(self):
        return self._callback

    @property
    def response_patterns(self):
        return ["/done", "/b_zero", self.buffer_id], None



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class Ball(UGen):
    """
    A bouncing ball physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ball = supriya.ugens.Ball.ar(
        ...     damping=0,
        ...     friction=0.01,
        ...     gravity=1,
        ...     source=source,
        ... )
        >>> ball
        Ball.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("gravity", 1.0), ("damping", 0.0), ("friction", 0.01)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Pluck(UGen):
    """
    A Karplus-String UGen.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(2)
        >>> pluck = supriya.ugens.Pluck.ar(
        ...     coefficient=0.5,
        ...     decay_time=1,
        ...     delay_time=0.2,
        ...     maximum_delay_time=0.2,
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> pluck
        Pluck.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("trigger", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1),
            ("coefficient", 0.5),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Spring(UGen):
    """
    A resonating spring physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> spring = supriya.ugens.Spring.ar(
        ...     damping=0,
        ...     source=source,
        ...     spring=1,
        ... )
        >>> spring
        Spring.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("spring", 1), ("damping", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TBall(UGen):
    """
    A bouncing object physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> tball = supriya.ugens.TBall.ar(
        ...     damping=0,
        ...     friction=0.01,
        ...     gravity=10,
        ...     source=source,
        ... )
        >>> tball
        TBall.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("gravity", 10), ("damping", 0), ("friction", 0.01)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import supriya.synthdefs
import supriya.ugens

"""
SynthDef("clap", {
    arg outBus=0, amp = 0.5;
    var env1, env2, out, noise1, noise2;

    env1 = EnvGen.ar(
        Env.new(
            [0, 1, 0, 1, 0, 1, 0, 1, 0],
            [0.001, 0.013, 0, 0.01, 0, 0.01, 0, 0.03],
            [0, -3, 0, -3, 0, -3, 0, -4],
            )
        );
    env2 = EnvGen.ar(
        Env.new(
            [0, 1, 0],
            [0.02, 0.3],
            [0, -4]
            ),
        doneAction:2,
        );

    noise1 = WhiteNoise.ar(env1);
    noise1 = HPF.ar(noise1, 600);
    noise1 = BPF.ar(noise1, 2000, 3);

    noise2 = WhiteNoise.ar(env2);
    noise2 = HPF.ar(noise2, 1000);
    noise2 = BPF.ar(noise2, 1200, 0.7, 0.7);

    out = noise1 + noise2;
    out = out * 2;
    out = out.softclip * amp;

    Out.ar(outBus, out.dup);
})
"""


def _build_clap_synthdef():

    with supriya.synthdefs.SynthDefBuilder(out=0, amplitude=0.5) as builder:

        envelope_one = supriya.synthdefs.Envelope(
            amplitudes=(0, 1, 0, 1, 0, 1, 0, 1, 0),
            durations=(0.001, 0.013, 0, 0.01, 0, 0.01, 0, 0.03),
            curves=(0, -3, 0, -3, 0, -3, 0, -4),
        )
        envelope_generator_one = supriya.ugens.EnvGen.ar(envelope=envelope_one)

        envelope_two = supriya.synthdefs.Envelope(
            amplitudes=(0, 1, 0), durations=(0.02, 0.3), curves=(0, -4)
        )
        envelope_generator_two = supriya.ugens.EnvGen.ar(
            envelope=envelope_two, done_action=2
        )

        noise_one = supriya.ugens.WhiteNoise.ar() * envelope_generator_one
        noise_one = supriya.ugens.HPF.ar(source=noise_one, frequency=600)
        noise_one = supriya.ugens.BPF.ar(
            source=noise_one, frequency=2000, reciprocal_of_q=3
        )

        noise_two = supriya.ugens.WhiteNoise.ar() * envelope_generator_two
        noise_two = supriya.ugens.HPF.ar(source=noise_two, frequency=1000)
        noise_two = supriya.ugens.BPF.ar(
            source=noise_two, frequency=1200, reciprocal_of_q=0.7
        )
        noise_two = noise_two * 0.7

        result = noise_one + noise_two
        result = result * 2
        result = result.softclip()
        result = result * builder["amplitude"]

        supriya.ugens.Out.ar(bus=builder["out"], source=(result, result))

    synthdef = builder.build()
    return synthdef


clap = _build_clap_synthdef()

__all__ = ("clap",)



import atexit
import logging
import queue
import threading
from typing import Optional, Tuple

from .bases import BaseClock
from .ephemera import Moment

logger = logging.getLogger("supriya.clocks")


class Clock(BaseClock):

    ### CLASS VARIABLES ###

    _default_clock = None

    ### INITIALIZER ###

    def __init__(self):
        BaseClock.__init__(self)
        self._event = threading.Event()
        self._thread = threading.Thread(target=self._run, daemon=True)
        atexit.register(self.stop)

    ### SCHEDULING METHODS ###

    def _enqueue_command(self, command):
        super()._enqueue_command(command)
        self._event.set()

    def _run(self, *args, offline=False, **kwargs):
        logger.debug(f"[{self.name}] Thread start")
        self._process_command_deque(first_run=True)
        while self._is_running:
            logger.debug(f"[{self.name}] Loop start")
            if not self._wait_for_queue():
                return
            try:
                current_moment = self._wait_for_moment()
            except queue.Empty:
                continue
            if current_moment is None:
                return
            current_moment = self._perform_events(current_moment)
            self._state = self._state._replace(
                previous_seconds=current_moment.seconds,
                previous_offset=current_moment.offset,
            )
            if not offline:
                self._event.wait(timeout=self._slop)
        logger.debug(f"[{self.name}] Terminating")

    def _wait_for_moment(self, offline=False) -> Optional[Moment]:
        current_time = self.get_current_time()
        next_time = self._event_queue.peek().seconds
        logger.debug(
            f"[{self.name}] ... Waiting for next moment at {next_time} from {current_time}"
        )
        while current_time < next_time:
            if not offline:
                self._event.wait(timeout=self._slop)
            if not self._is_running:
                return None
            self._process_command_deque()
            next_time = self._event_queue.peek().seconds
            current_time = self.get_current_time()
            self._event.clear()
        return self._seconds_to_moment(current_time)

    def _wait_for_queue(self, offline=False) -> bool:
        logger.debug(f"[{self.name}] ... Waiting for events")
        self._process_command_deque()
        self._event.clear()
        while not self._event_queue.qsize():
            if not offline:
                self._event.wait(timeout=self._slop)
            if not self._is_running:
                return False
            self._process_command_deque()
            self._event.clear()
        return True

    ### PUBLIC METHODS ###

    def cancel(self, event_id) -> Optional[Tuple]:
        event = super().cancel(event_id)
        self._event.set()
        return event

    @classmethod
    def default(cls):
        if cls._default_clock is None:
            cls._default_clock = cls()
        return cls._default_clock

    def start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ):
        self._start(
            initial_time=initial_time,
            initial_offset=initial_offset,
            initial_measure=initial_measure,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
        )
        self._thread = threading.Thread(target=self._run, args=(self,), daemon=True)
        self._thread.start()

    def stop(self):
        if self._stop():
            self._event.set()
            self._thread.join()



# distutils: language = c++

from shm cimport server_shared_memory_client


cdef class ServerSHM:
    cdef server_shared_memory_client* client
    cdef unsigned int bus_count

    def __cinit__(self, unsigned int port_number, unsigned int bus_count):
        self.client = new server_shared_memory_client(port_number)
        self.bus_count = bus_count

    def __dealloc__(self):
        del self.client

    def __getitem__(self, item):
        if isinstance(item, int):
            if item < 0 or item >= self.bus_count:
                raise ValueError("index out of bounds")
            return self.client.get_control_busses()[item]
        elif isinstance(item, slice):
            result = []
            for index in range(*item.indices(self.bus_count)):
                result.append(self.client.get_control_busses()[index])
            return result
        raise ValueError(item)



import abc
import functools

from uqbar.objects import get_repr

from supriya.system import SupriyaObject


class SessionObject(SupriyaObject):
    """
    A non-realtime session object, analogous to ServerObject.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, session):
        import supriya.nonrealtime

        prototype = (supriya.nonrealtime.Session, type(None))
        assert isinstance(session, prototype)
        self._session = session

    ### SPECIAL METHODS ###

    def __repr__(self):
        return get_repr(self, multiline=False)

    ### PUBLIC METHODS ###

    @staticmethod
    def require_offset(function):
        @functools.wraps(function)
        def wrapper(self, *args, **kwargs):
            import supriya.nonrealtime

            if isinstance(self, supriya.nonrealtime.Session):
                session = self
            else:
                session = self.session
            if "offset" not in kwargs or kwargs["offset"] is None:
                if not session._active_moments:
                    raise ValueError("No active moment.")
                offset = session._active_moments[-1].offset
                kwargs["offset"] = offset
            if isinstance(self, SessionObject):
                if not (self.start_offset <= kwargs["offset"] <= self.stop_offset):
                    raise ValueError(
                        "Offset {} must intersect [{}, {}]".format(
                            float(offset), self.start_offset, self.stop_offset
                        )
                    )
            with session.at(kwargs["offset"]):
                return function(self, *args, **kwargs)

        return wrapper

    ### PUBLIC PROPERTIES ###

    @property
    def session(self):
        return self._session



import collections
from collections.abc import Sequence

from supriya import CalculationRate, ParameterRate, SignalRange
from supriya.system import SupriyaValueObject
from supriya.typing import UGenInputMap

from .bases import MultiOutUGen
from .mixins import UGenMethodMixin


class Range(SupriyaValueObject):
    """
    A range.

    ::

        >>> supriya.synthdefs.Range(-1.0, 1.0)
        Range(
            maximum=1.0,
            minimum=-1.0,
        )

    ::

        >>> supriya.synthdefs.Range(minimum=0.0)
        Range(
            maximum=inf,
            minimum=0.0,
        )

    ::

        >>> supriya.synthdefs.Range()
        Range(
            maximum=inf,
            minimum=-inf,
        )

    ::

        >>> supriya.synthdefs.Range((0.1, 0.9))
        Range(
            maximum=0.9,
            minimum=0.1,
        )

    ::

        >>> supriya.synthdefs.Range(supriya.synthdefs.Range(-3, 3))
        Range(
            maximum=3.0,
            minimum=-3.0,
        )

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_minimum", "_maximum")

    ### INITIALIZER ###

    def __init__(self, minimum=None, maximum=None):
        if isinstance(minimum, Sequence) and maximum is None and len(minimum) == 2:
            minimum, maximum = minimum
        elif isinstance(minimum, type(self)):
            minimum, maximum = minimum.minimum, minimum.maximum
        if minimum is None:
            minimum = float("-inf")
        if not isinstance(minimum, (float, int)):
            raise ValueError(minimum)
        minimum = float(minimum)
        if maximum is None:
            maximum = float("inf")
        if not isinstance(maximum, (float, int)):
            raise ValueError(maximum)
        maximum = float(maximum)
        assert minimum <= maximum
        self._minimum = minimum
        self._maximum = maximum

    ### PUBLIC METHODS ###

    @staticmethod
    def scale(value, input_range, output_range, exponent=1.0):
        """
        Scales `value` from `input_range` to `output_range`.

        Curve value exponentially by `exponent`.

        ::

            >>> input_range = supriya.synthdefs.Range(0.0, 10.0)
            >>> output_range = supriya.synthdefs.Range(-2.5, 2.5)

        ::

            >>> supriya.synthdefs.Range.scale(0.0, input_range, output_range)
            -2.5

        ::

            >>> supriya.synthdefs.Range.scale(5.0, input_range, output_range)
            0.0

        ::

            >>> supriya.synthdefs.Range.scale(5.0, input_range, output_range, 2.0)
            -1.25

        ::

            >>> supriya.synthdefs.Range.scale(5.0, input_range, output_range, 0.5)
            1.0355...

        Returns float.
        """
        value = (value - input_range.minimum) / input_range.width
        if exponent != 1:
            value = pow(value, exponent)
        value = (value * output_range.width) + output_range.minimum
        return value

    ### PUBLIC PROPERTIES ###

    @property
    def maximum(self):
        return self._maximum

    @property
    def minimum(self):
        return self._minimum

    @property
    def width(self):
        return self._maximum - self._minimum


class Parameter(UGenMethodMixin, SupriyaValueObject):

    ### CLASS VARIABLES ###

    __slots__ = (
        "_lag",
        "_name",
        "_parameter_rate",
        "_range",
        "_unit",
        "_uuid",
        "_value",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        lag=None,
        name=None,
        parameter_rate=ParameterRate.CONTROL,
        range_=None,
        unit=None,
        value=None,
    ):
        if lag is not None:
            lag = float(lag)
        self._lag = lag
        if name is not None:
            name = str(name)
        self._name = name
        self._parameter_rate = ParameterRate.from_expr(parameter_rate)
        if range_ is not None:
            assert isinstance(range_, Range)
        self._range = range_
        self._unit = unit
        self._uuid = None
        if isinstance(value, Sequence):
            value = tuple(float(_) for _ in value)
            assert value, value
        else:
            value = float(value)
        self._value = value

    ### SPECIAL METHODS ###

    def __getitem__(self, i):
        return self._get_output_proxy(i)

    def __len__(self):
        if isinstance(self.value, float):
            return 1
        return len(self.value)

    ### PRIVATE METHODS ###

    def _get_source(self):
        return self

    def _get_output_number(self):
        return 0

    def _optimize_graph(self, sort_bundles):
        pass

    ### PUBLIC PROPERTIES ###

    @property
    def calculation_rate(self):
        return CalculationRate.from_expr(self)

    @property
    def has_done_flag(self):
        return False

    @property
    def inputs(self):
        return ()

    @property
    def lag(self):
        return self._lag

    @property
    def name(self):
        return self._name

    @property
    def parameter_rate(self):
        return self._parameter_rate

    @property
    def range_(self):
        return self._range

    @property
    def signal_range(self):
        SignalRange.BIPOLAR

    @property
    def unit(self):
        return self._unit

    @property
    def value(self):
        return self._value


class Control(MultiOutUGen):
    """
    A control-rate control ugen.

    Control ugens can be set and routed externally to interact with a running
    synth. Controls are created from the parameters of a synthesizer
    definition, and typically do not need to be created by hand.
    """

    ### CLASS VARIABLES ###

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])

    __slots__ = ("_parameters",)

    ### INITIALIZER ###

    def __init__(self, parameters, calculation_rate=None, starting_control_index=0):
        coerced_parameters = []
        for parameter in parameters:
            if not isinstance(parameter, Parameter):
                parameter = Parameter(name=parameter, value=0)
            coerced_parameters.append(parameter)
        self._parameters = tuple(coerced_parameters)
        MultiOutUGen.__init__(
            self,
            channel_count=len(self),
            calculation_rate=calculation_rate,
            special_index=starting_control_index,
        )

    ### SPECIAL METHODS ###

    def __getitem__(self, i):
        """
        Gets output proxy at `i`, via index or control name.

        Returns output proxy.
        """
        import supriya.synthdefs

        if type(i) == int:
            if len(self) == 1:
                return supriya.synthdefs.OutputProxy(self, 0)
            return supriya.synthdefs.OutputProxy(self, i)
        else:
            return self[self._get_control_index(i)]

    def __len__(self):
        """
        Gets number of ugen outputs.

        Equal to the number of control names.

        Returns integer.
        """
        return sum(len(_) for _ in self.parameters)

    ### PRIVATE METHODS ###

    def _get_control_index(self, control_name):
        for i, parameter in enumerate(self._parameters):
            if parameter.name == control_name:
                return i
        raise ValueError

    def _get_outputs(self):
        return [self.calculation_rate] * len(self)

    def _get_parameter_output_proxies(self):
        output_proxies = []
        for parameter in self.parameters:
            output_proxies.extend(parameter)
        return output_proxies

    ### PUBLIC PROPERTIES ###

    @property
    def controls(self):
        """
        Gets controls of control ugen.

        Returns ugen graph.
        """
        import supriya.synthdefs

        if len(self.parameters) == 1:
            result = self
        else:
            result = [
                supriya.synthdefs.OutputProxy(self, i)
                for i in range(len(self.parameters))
            ]
        return result

    @property
    def parameters(self):
        """
        Gets control names associated with control.

        Returns tuple.
        """
        return self._parameters

    @property
    def starting_control_index(self):
        """
        Gets starting control index of control ugen.

        Equivalent to the control ugen's special index.

        Returns integer.
        """
        return self._special_index


class AudioControl(Control):
    """
    A trigger-rate control ugen.
    """

    def __init__(self, parameters, calculation_rate=None, starting_control_index=0):
        Control.__init__(
            self,
            parameters,
            calculation_rate=CalculationRate.AUDIO,
            starting_control_index=starting_control_index,
        )


class LagControl(Control):
    """
    A lagged control-rate control ugen.
    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("lags", None)])

    _unexpanded_input_names = ("lags",)

    ### INITIALIZER ###

    def __init__(self, parameters, calculation_rate=None, starting_control_index=0):
        coerced_parameters = []
        for parameter in parameters:
            if not isinstance(parameter, Parameter):
                parameter = Parameter(name=parameter, value=0)
            coerced_parameters.append(parameter)
        self._parameters = tuple(coerced_parameters)
        lags = []
        for parameter in self._parameters:
            lag = parameter.lag or 0.0
            lags.extend([lag] * len(parameter))
        MultiOutUGen.__init__(
            self,
            channel_count=len(self),
            calculation_rate=calculation_rate,
            special_index=starting_control_index,
            lags=lags,
        )


class TrigControl(Control):
    """
    A trigger-rate control ugen.
    """

    ### CLASS VARIABLES ###

    ### INITIALIZER ##

    def __init__(self, parameters, calculation_rate=None, starting_control_index=0):
        Control.__init__(
            self,
            parameters,
            calculation_rate=CalculationRate.CONTROL,
            starting_control_index=starting_control_index,
        )



import os
import pathlib
import platform
import signal
import subprocess
from dataclasses import dataclass
from typing import Optional

import uqbar.io
import uqbar.objects

import supriya


@dataclass(frozen=True)
class Options:
    """
    SuperCollider server options configuration.

    ::

        >>> import supriya.realtime
        >>> options = supriya.scsynth.Options()

    """

    ### CLASS VARIABLES ###

    audio_bus_channel_count: int = 1024
    block_size: int = 64
    buffer_count: int = 1024
    control_bus_channel_count: int = 16384
    hardware_buffer_size: Optional[int] = None
    initial_node_id: int = 1000
    input_bus_channel_count: int = 8
    input_device: Optional[str] = None
    input_stream_mask: str = ""
    load_synthdefs: bool = True
    maximum_logins: int = 1
    maximum_node_count: int = 1024
    maximum_synthdef_count: int = 1024
    memory_locking: bool = False
    memory_size: int = 8192
    output_bus_channel_count: int = 8
    output_device: Optional[str] = None
    output_stream_mask: str = ""
    password: Optional[str] = None
    protocol: str = "udp"
    random_number_generator_count: int = 64
    remote_control_volume: bool = False
    restricted_path: Optional[str] = None
    sample_rate: Optional[int] = None
    threads: Optional[int] = None
    ugen_plugins_path: Optional[str] = None
    verbosity: int = 0
    wire_buffer_count: int = 64
    zero_configuration: bool = False

    ### INITIALIZER ###

    def __post_init__(self):
        if self.input_bus_channel_count is None:
            object.__setattr__(self, "input_bus_channel_count", 8)
        if self.output_bus_channel_count is None:
            object.__setattr__(self, "output_bus_channel_count", 8)
        if self.input_bus_channel_count < 0:
            raise ValueError(self.input_bus_channel_count)
        if self.output_bus_channel_count < 0:
            raise ValueError(self.output_bus_channel_count)
        if self.audio_bus_channel_count < (
            self.input_bus_channel_count + self.output_bus_channel_count
        ):
            raise ValueError("Insufficient audio buses")

    ### CLASS VARIABLES ###

    def __repr__(self):
        return uqbar.objects.get_repr(self, multiline=True, suppress_defaults=False)

    ### PUBLIC METHODS ###

    def as_options_string(self, port=57110, realtime=True, supernova=False):
        result = []
        if realtime:
            if self.protocol == "tcp":
                result.append("-t {}".format(port))
            else:
                result.append("-u {}".format(port))
            if self.input_device:
                flag = "-H {}".format(self.input_device)
                if self.output_device != self.input_device:
                    flag = "{} {}".format(flag, self.output_device)
                result.append(flag)
            if self.maximum_logins != 64:
                result.append("-l {}".format(self.maximum_logins))
            if self.password:
                result.append("-p {}".format(self.password))
            if self.sample_rate is not None:
                result.append("-S {}".format(int(self.sample_rate)))
            if not self.zero_configuration:
                result.append("-R 0")
        if self.audio_bus_channel_count != 1024:
            result.append("-a {}".format(self.audio_bus_channel_count))
        if self.control_bus_channel_count != 16384:
            result.append("-c {}".format(self.control_bus_channel_count))
        if self.input_bus_channel_count != 8:
            result.append("-i {}".format(self.input_bus_channel_count))
        if self.output_bus_channel_count != 8:
            result.append("-o {}".format(self.output_bus_channel_count))
        if self.buffer_count != 1024:
            result.append("-b {}".format(self.buffer_count))
        if self.maximum_node_count != 1024:
            result.append("-n {}".format(self.maximum_node_count))
        if self.maximum_synthdef_count != 1024:
            result.append("-d {}".format(self.maximum_synthdef_count))
        if self.block_size != 64:
            result.append("-z {}".format(self.block_size))
        if self.hardware_buffer_size is not None:
            result.append("-Z {}".format(int(self.hardware_buffer_size)))
        if self.memory_size != 8192:
            result.append("-m {}".format(self.memory_size))
        if self.random_number_generator_count != 64:
            result.append("-r {}".format(self.random_number_generator_count))
        if self.wire_buffer_count != 64:
            result.append("-w {}".format(self.wire_buffer_count))
        if not self.load_synthdefs:
            result.append("-D 0")
        if self.input_stream_mask:
            result.append("-I {}".format(self.input_stream_mask))
        if self.output_stream_mask:
            result.append("-O {}".format(self.output_stream_mask))
        if 0 < self.verbosity:
            result.append("-v {}".format(self.verbosity))
        if self.restricted_path is not None:
            result.append("-P {}".format(self.restricted_path))
        if self.memory_locking:
            result.append("-L")
        if self.ugen_plugins_path:
            result.append("-U {}".format(self.ugen_plugins_path))
        if supernova and self.threads:
            result.append("-t {}".format(self.threads))
        options_string = " ".join(sorted(result))
        return options_string

    ### PUBLIC PROPERTIES ###

    @property
    def first_private_bus_id(self):
        return self.output_bus_channel_count + self.input_bus_channel_count

    @property
    def private_audio_bus_channel_count(self):
        return (
            self.audio_bus_channel_count
            - self.input_bus_channel_count
            - self.output_bus_channel_count
        )


def _fallback_scsynth_path():
    if platform.system() == "Darwin":
        for path in [
            pathlib.Path("/Applications/SuperCollider.app/Contents/Resources/scsynth"),
            pathlib.Path(
                "/Applications/SuperCollider/SuperCollider.app/Contents/Resources/scsynth"
            ),
        ]:
            if path.exists():
                return path
    elif platform.system() == "Linux":
        for path in [
            pathlib.Path("/usr/bin/scsynth"),
            pathlib.Path("/usr/local/bin/scsynth"),
        ]:
            if path.exists():
                return path
    return None


def find(scsynth_path=None):
    """Find the ``scsynth`` executable.

    The following paths, if defined, will be searched (prioritised as ordered):
    1. The absolute path ``scsynth_path``
    2. The environment variable ``SCSYNTH_PATH``
    3. ``scsynth_path`` if defined in Supriya's configuration file
    4. The user's ``PATH``
    5. Common installation directories of the SuperCollider application.

    Returns a path to the ``scsynth`` executable.
    Raises ``RuntimeError`` if no path is found.
    """
    scsynth_path = pathlib.Path(
        scsynth_path
        or os.environ.get("SCSYNTH_PATH")
        or supriya.config.get("core", "scsynth_path")
        or "scsynth"
    )
    if scsynth_path.is_absolute() and uqbar.io.find_executable(scsynth_path):
        return scsynth_path
    scsynth_path_candidates = uqbar.io.find_executable(scsynth_path.name)
    if scsynth_path_candidates:
        return pathlib.Path(scsynth_path_candidates[0])
    fallback_path = _fallback_scsynth_path()
    if fallback_path is not None:
        return fallback_path
    raise RuntimeError("Failed to locate scsynth")


def kill(supernova=False):
    executable = "supernova" if supernova else "scsynth"
    with subprocess.Popen(
        ["ps", "-Af"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    ) as process:
        output = process.stdout.read()
    for line in output.decode().splitlines():
        parts = line.split()
        if not any(part == executable for part in parts):
            continue
        pid = int(parts[1])
        os.kill(pid, signal.SIGKILL)



import abc
import asyncio
import logging
import threading
import time
from collections import deque

from uqbar.objects import new

from supriya.exceptions import ServerOffline
from supriya.osc.messages import BUNDLE_PREFIX, OscBundle
from supriya.system import SupriyaValueObject

logger = logging.getLogger("supriya.osc")


class Requestable(SupriyaValueObject):

    ### INITIALIZER ###

    def __init__(self):
        self._condition = threading.Condition()
        self._response = None

    ### PRIVATE METHODS ###

    def _get_response_patterns_and_requestable(self, server):
        raise NotImplementedError

    def _handle_async(self, sync, server):
        raise NotImplementedError

    def _linearize(self):
        raise NotImplementedError

    def _sanitize_node_id(self, node_id, with_placeholders):
        if not isinstance(node_id, int) and with_placeholders:
            return -1
        return int(node_id)

    def _set_response(self, message):
        from supriya.commands import Response

        with self.condition:
            self._response = Response.from_osc_message(message)
            self.condition.notify()

    def _set_response_async(self, message):
        from supriya.commands import Response

        self._response = Response.from_osc_message(message)
        self._response_future.set_result(True)

    ### PUBLIC METHODS ###

    def communicate(self, server, sync=True, timeout=1.0, apply_local=True):
        import supriya.realtime

        if not isinstance(server, supriya.realtime.servers.BaseServer):
            raise ValueError(server)
        if not server.is_running:
            raise ServerOffline
        if apply_local:
            with server._lock:
                for request in self._linearize():
                    request._apply_local(server)
        # handle non-sync
        if self._handle_async(sync, server):
            return
        (
            success_pattern,
            failure_pattern,
            requestable,
        ) = self._get_response_patterns_and_requestable(server)
        start_time = time.time()
        timed_out = False
        with self.condition:
            try:
                server.osc_protocol.register(
                    pattern=success_pattern,
                    failure_pattern=failure_pattern,
                    procedure=self._set_response,
                    once=True,
                )
            except Exception:
                print(self)
                raise
            server.send(requestable.to_osc())
            while self.response is None:
                self.condition.wait(timeout)
                current_time = time.time()
                delta_time = current_time - start_time
                if timeout <= delta_time:
                    timed_out = True
                    break
        if timed_out:
            logger.warning("Timed out: {!r}".format(self))
            return None
        return self._response

    async def communicate_async(self, server, sync=True, timeout=1.0):
        (
            success_pattern,
            failure_pattern,
            requestable,
        ) = self._get_response_patterns_and_requestable(server)
        if self._handle_async(sync, server):
            return
        loop = asyncio.get_running_loop()
        self._response_future = loop.create_future()
        server.osc_protocol.register(
            pattern=success_pattern,
            failure_pattern=failure_pattern,
            procedure=self._set_response_async,
            once=True,
        )
        server.send(requestable.to_osc())
        await asyncio.wait_for(self._response_future, timeout=timeout)
        return self._response

    def to_datagram(self, *, with_placeholders=False):
        return self.to_osc(with_placeholders=with_placeholders).to_datagram()

    def to_list(self, *, with_placeholders=False):
        return self.to_osc(with_placeholders=with_placeholders).to_list()

    ### PUBLIC PROPERTIES ###

    @property
    def condition(self):
        return self._condition

    @property
    def response(self):
        return self._response


class Request(Requestable):

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        pass

    def _get_response_patterns_and_requestable(self, server):
        success_pattern, failure_pattern = self.response_patterns
        return success_pattern, failure_pattern, self

    def _handle_async(self, sync, server):
        if not sync or self.response_patterns[0] is None:
            message = self.to_osc()
            server.send(message)
            return True

    def _linearize(self):
        if hasattr(self, "callback") and self.callback:
            yield new(self, callback=None)
            yield from self.callback._linearize()
        else:
            yield self

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def to_osc(self, *, with_placeholders=False):
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def request_name(self):
        return self.request_id.request_name

    @property
    def response_patterns(self):
        return None, None

    @property
    @abc.abstractmethod
    def request_id(self):
        return NotImplementedError


class RequestBundle(Requestable):
    """
    A Request bundle.

    ::

        >>> request_one = supriya.commands.BufferAllocateRequest(
        ...     buffer_id=23,
        ...     frame_count=512,
        ...     channel_count=1,
        ... )
        >>> request_two = supriya.commands.BufferAllocateRequest(
        ...     buffer_id=24,
        ...     frame_count=512,
        ...     channel_count=1,
        ... )
        >>> request_bundle = supriya.commands.RequestBundle(
        ...     timestamp=10.5,
        ...     contents=[request_one, request_two],
        ... )

    ::

        >>> request_bundle.to_osc()
        OscBundle(
            contents=(
                OscMessage('/b_alloc', 23, 512, 1),
                OscMessage('/b_alloc', 24, 512, 1),
            ),
            timestamp=10.5,
        )

    ::

        >>> request_bundle.to_list()
        [10.5, [['/b_alloc', 23, 512, 1], ['/b_alloc', 24, 512, 1]]]

    """

    ### INITIALIZER ###

    def __init__(self, timestamp=None, contents=None):
        import supriya.commands

        self._condition = threading.Condition()
        self._timestamp = timestamp
        if contents is not None:
            prototype = (supriya.commands.Request, type(self))
            assert all(isinstance(x, prototype) for x in contents)
            contents = tuple(contents)
        else:
            contents = ()
        self._contents = contents
        self._response = None

    ### PRIVATE METHODS ###

    def _get_response_patterns_and_requestable(self, server):
        from supriya.commands import SyncRequest

        sync_id = server.next_sync_id
        contents = list(self.contents)
        contents.append(SyncRequest(sync_id=sync_id))
        request_bundle = type(self)(contents=contents)
        response_pattern = ["/synced", sync_id]
        return response_pattern, None, request_bundle

    def _handle_async(self, sync, server):
        if not sync:
            message = self.to_osc()
            server.send(message)
            return True

    def _linearize(self):
        for x in self.contents:
            yield from x._linearize()

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        contents = []
        for x in self.contents:
            if isinstance(x, type(self)):
                contents.append(x.to_osc(with_placeholders=with_placeholders))
            else:
                contents.append(x.to_osc(with_placeholders=with_placeholders))
        bundle = OscBundle(timestamp=self.timestamp, contents=contents)
        return bundle

    @classmethod
    def partition(cls, requests, timestamp=None):
        bundles = []
        contents = []
        requests = deque(requests)
        remaining = maximum = 8192 - len(BUNDLE_PREFIX) - 4
        while requests:
            request = requests.popleft()
            datagram = request.to_datagram()
            remaining -= len(datagram) + 4
            if remaining > 0:
                contents.append(request)
            else:
                bundles.append(cls(timestamp=timestamp, contents=contents))
                contents = [request]
                remaining = maximum
        if contents:
            bundles.append(cls(timestamp=timestamp, contents=contents))
        return bundles

    ### PUBLIC PROPERTIES ###

    @property
    def condition(self):
        return self._condition

    @property
    def contents(self):
        return self._contents

    @property
    def timestamp(self):
        return self._timestamp


class Response(SupriyaValueObject):

    ### PRIVATE METHODS ###

    @staticmethod
    def _group_items(items, length):
        iterators = [iter(items)] * length
        iterator = zip(*iterators)
        return iterator

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, message):
        import supriya.commands

        return {
            "/b_info": supriya.commands.BufferInfoResponse,
            "/b_set": supriya.commands.BufferSetResponse,
            "/b_setn": supriya.commands.BufferSetContiguousResponse,
            "/c_set": supriya.commands.ControlBusSetResponse,
            "/c_setn": supriya.commands.ControlBusSetContiguousResponse,
            "/d_removed": supriya.commands.SynthDefRemovedResponse,
            "/done": supriya.commands.DoneResponse,
            "/fail": supriya.commands.FailResponse,
            "/g_queryTree.reply": supriya.commands.QueryTreeResponse,
            "/n_end": supriya.commands.NodeInfoResponse,
            "/n_go": supriya.commands.NodeInfoResponse,
            "/n_info": supriya.commands.NodeInfoResponse,
            "/n_move": supriya.commands.NodeInfoResponse,
            "/n_off": supriya.commands.NodeInfoResponse,
            "/n_on": supriya.commands.NodeInfoResponse,
            "/n_set": supriya.commands.NodeSetResponse,
            "/n_setn": supriya.commands.NodeSetContiguousResponse,
            "/status.reply": supriya.commands.StatusResponse,
            "/synced": supriya.commands.SyncedResponse,
            "/tr": supriya.commands.TriggerResponse,
        }[message.address].from_osc_message(message)

    def to_dict(self):
        result = {}
        for key, value in self.__getstate__().items():
            key = key[1:]
            if key == "osc_message":
                continue
            result[key] = value
        return result



import collections

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen, WidthFirstUGen


class BufRd(MultiOutUGen):
    """
    A buffer-reading oscillator.

    ::

        >>> buffer_id = 23
        >>> phase = supriya.ugens.Phasor.ar(
        ...     rate=supriya.ugens.BufRateScale.kr(buffer_id),
        ...     start=0,
        ...     stop=supriya.ugens.BufFrames.kr(buffer_id),
        ... )
        >>> buf_rd = supriya.ugens.BufRd.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     interpolation=2,
        ...     loop=1,
        ...     phase=phase,
        ... )
        >>> buf_rd
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("phase", 0.0), ("loop", 1.0), ("interpolation", 2.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufWr(UGen):
    """
    A buffer-writing oscillator.

    ::

        >>> buffer_id = 23
        >>> phase = supriya.ugens.Phasor.ar(
        ...     rate=supriya.ugens.BufRateScale.kr(buffer_id),
        ...     start=0,
        ...     stop=supriya.ugens.BufFrames.kr(buffer_id),
        ... )
        >>> source = supriya.ugens.SoundIn.ar(bus=(0, 1))
        >>> buf_wr = supriya.ugens.BufWr.ar(
        ...     buffer_id=buffer_id,
        ...     loop=1,
        ...     phase=phase,
        ...     source=source,
        ... )
        >>> buf_wr
        BufWr.ar()

    """

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("phase", 0.0), ("loop", 1.0), ("source", None)]
    )
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class ClearBuf(WidthFirstUGen):
    """

    ::

        >>> clear_buf = supriya.ugens.ClearBuf.ir(
        ...     buffer_id=23,
        ... )
        >>> clear_buf
        ClearBuf.ir()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class LocalBuf(WidthFirstUGen):
    """
    A synth-local buffer.

    ::

        >>> local_buf = supriya.ugens.LocalBuf(
        ...     channel_count=1,
        ...     frame_count=1,
        ... )
        >>> local_buf
        LocalBuf.ir()

    LocalBuf creates a ``MaxLocalBufs`` UGen implicitly during SynthDef
    compilation:

    ::

        >>> with supriya.synthdefs.SynthDefBuilder() as builder:
        ...     local_buf = supriya.ugens.LocalBuf(2048)
        ...     source = supriya.ugens.PinkNoise.ar()
        ...     pv_chain = supriya.ugens.FFT(
        ...         buffer_id=local_buf,
        ...         source=source,
        ...     )
        ...     ifft = supriya.ugens.IFFT.ar(pv_chain=pv_chain)
        ...     out = supriya.ugens.Out.ar(bus=0, source=ifft)
        ...
        >>> synthdef = builder.build()
        >>> for ugen in synthdef.ugens:
        ...     ugen
        ...
        MaxLocalBufs.ir()
        LocalBuf.ir()
        PinkNoise.ar()
        FFT.kr()
        IFFT.ar()
        Out.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("channel_count", 1), ("frame_count", 1)]
    )
    _valid_calculation_rates = (CalculationRate.SCALAR,)

    ### INITIALIZER ###

    def __init__(self, frame_count=1, channel_count=1, calculation_rate=None):
        import supriya.synthdefs

        if calculation_rate is None:
            calculation_rate = supriya.CalculationRate.SCALAR
        WidthFirstUGen.__init__(
            self,
            calculation_rate=calculation_rate,
            channel_count=channel_count,
            frame_count=frame_count,
        )


class MaxLocalBufs(UGen):
    """
    Sets the maximum number of local buffers in a synth.

    Used internally by LocalBuf.

    ::

        >>> max_local_bufs = supriya.ugens.MaxLocalBufs(1)
        >>> max_local_bufs
        MaxLocalBufs.ir()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("maximum", 0)])
    _valid_calculation_rates = (CalculationRate.SCALAR,)

    ### INITIALIZER ###

    def __init__(self, maximum=0):
        import supriya.synthdefs

        maximum = float(maximum)
        calculation_rate = supriya.CalculationRate.SCALAR
        UGen.__init__(self, calculation_rate=calculation_rate, maximum=maximum)

    ### PUBLIC METHODS ###

    def increment(self):
        """
        Increments maximum local buffer count.

        ::

            >>> max_local_bufs = supriya.ugens.MaxLocalBufs(1)
            >>> max_local_bufs.maximum
            1.0

        ::

            >>> max_local_bufs.increment()
            >>> max_local_bufs.maximum
            2.0

        Returns none.
        """
        self._inputs[0] += 1


class PlayBuf(MultiOutUGen):
    """
    A sample playback oscillator.

    ::

        >>> buffer_id = 23
        >>> play_buf = supriya.ugens.PlayBuf.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     done_action=0,
        ...     loop=0,
        ...     rate=1,
        ...     start_position=0,
        ...     trigger=1,
        ... )
        >>> play_buf
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("rate", 1),
            ("trigger", 1),
            ("start_position", 0),
            ("loop", 0),
            ("done_action", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class RecordBuf(UGen):
    """
    Records or overdubs into a buffer.

    ::

        >>> buffer_id = 23
        >>> source = supriya.ugens.SoundIn.ar(bus=(0, 1))
        >>> record_buf = supriya.ugens.RecordBuf.ar(
        ...     buffer_id=buffer_id,
        ...     done_action=0,
        ...     loop=1,
        ...     offset=0,
        ...     preexisting_level=0,
        ...     record_level=1,
        ...     run=1,
        ...     source=source,
        ...     trigger=1,
        ... )
        >>> record_buf
        RecordBuf.ar()

    """

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("offset", 0),
            ("record_level", 1),
            ("preexisting_level", 0),
            ("run", 1),
            ("loop", 1),
            ("trigger", 1),
            ("done_action", 0),
            ("source", None),
        ]
    )
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



"""
Version information.

- Major version is last two digits of the current year.
- Minor version is the month as a digit.
- Patch / beta tag is the number of releases during that month.

This follows black's versioning scheme.
"""
__version_info__ = (22, "2b0")
__version__ = ".".join(str(x) for x in __version_info__)



"""
Utility functions.

These will be migrated out into a base package at some point.
"""

import importlib
import itertools
import pathlib
from collections.abc import Iterable, Sequence


def expand(mapping, unexpanded=None, only=None):
    expanded_mappings = []
    maximum_length = 1
    massaged = {}
    for key, value in mapping.items():
        if only and key not in only:
            continue
        if isinstance(value, Sequence):
            if key not in (unexpanded or ()):
                maximum_length = max(len(value), maximum_length)
        else:
            value = [value]
        massaged[key] = value
    for i in range(maximum_length):
        expanded_mapping = {}
        for key, value in massaged.items():
            if key in (unexpanded or ()):
                expanded_mapping[key] = value
            else:
                expanded_mapping[key] = value[i % len(value)]
        expanded_mappings.append(expanded_mapping)
    return expanded_mappings


def flatten_iterable(iterable):
    for x in iterable:
        if isinstance(x, Iterable):
            yield from flatten_iterable(x)
        else:
            yield x


def group_iterable_by_count(iterable, count):
    iterator = iter(iterable)
    while True:
        group = list(itertools.islice(iterator, count))
        if not group:
            return
        yield group


def iterate_nwise(iterable, n=2):
    iterables = itertools.tee(iterable, n)
    temp = []
    for idx, it in enumerate(iterables):
        it = itertools.islice(it, idx, None)
        temp.append(it)
    return zip(*temp)


def locate(path: str) -> pathlib.Path:
    if ":" in path:
        module_path, _, file_path = path.partition(":")
        module = importlib.import_module(module_path)
        if hasattr(module, "__file__") and module.__file__ is not None:
            return pathlib.Path(module.__file__).parent / file_path
        return pathlib.Path(module.__path__[0]) / file_path  # type: ignore
    return pathlib.Path(path)


def repeat_sequence_to_length(sequence, length):
    for i, x in enumerate(itertools.cycle(sequence)):
        if i >= length:
            break
        yield x


def zip_sequences(*args):
    maximum_i = max(len(_) for _ in args) - 1
    cycles = [itertools.cycle(_) for _ in args]
    iterator = enumerate(zip(*cycles))
    for i, result in iterator:
        yield result
        if i == maximum_i:
            break



"""
Tools for constructing and compiling synthesizer definitions (SynthDefs).
"""
from .mixins import OutputProxy, UGenArray, UGenMethodMixin  # isort:skip
from .bases import (
    BinaryOpUGen,
    MultiOutUGen,
    PseudoUGen,
    PureMultiOutUGen,
    PureUGen,
    UGen,
    UGenMeta,
    UnaryOpUGen,
    WidthFirstUGen,
)
from .builders import SynthDefBuilder, synthdef
from .compilers import SynthDefCompiler, SynthDefDecompiler
from .controls import AudioControl, Control, LagControl, Parameter, Range, TrigControl
from .envelopes import Envelope
from .factories import SynthDefFactory
from .grapher import SynthDefGrapher
from .synthdefs import SuperColliderSynthDef, SynthDef, UGenSortBundle

__all__ = [
    "AudioControl",
    "BinaryOpUGen",
    "Control",
    "Envelope",
    "LagControl",
    "MultiOutUGen",
    "OutputProxy",
    "Parameter",
    "PseudoUGen",
    "PureMultiOutUGen",
    "PureUGen",
    "Range",
    "SuperColliderSynthDef",
    "SynthDef",
    "SynthDefBuilder",
    "SynthDefCompiler",
    "SynthDefDecompiler",
    "SynthDefFactory",
    "SynthDefGrapher",
    "TrigControl",
    "UGen",
    "UGenArray",
    "UGenMeta",
    "UGenMethodMixin",
    "UGenSortBundle",
    "UnaryOpUGen",
    "WidthFirstUGen",
    "synthdef",
]



import bisect
import collections
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import uqbar.graphs
from uqbar.objects import new

import supriya  # noqa
from supriya.commands import GroupNewRequest, SynthNewRequest
from supriya.enums import AddAction, ParameterRate
from supriya.nonrealtime.bases import SessionObject
from supriya.nonrealtime.states import NodeTransition, State
from supriya.synthdefs import SynthDef
from supriya.typing import AddActionLike


class Node(SessionObject):
    """
    A non-realtime node.
    """

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = ()

    ### INITIALIZER ###

    def __init__(
        self,
        session: "supriya.nonrealtime.Session",
        session_id: int,
        duration: Optional[float] = None,
        start_offset: Optional[float] = None,
    ) -> None:
        SessionObject.__init__(self, session)
        self._session_id = int(session_id)
        start_offset = start_offset or 0
        self._start_offset = float(start_offset)
        if duration is None:
            duration = float("inf")
        self._duration = duration
        self._events: Dict[str, List[Tuple[float, float]]] = {}

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        return "<{} #{} @{}:{}>".format(
            type(self).__name__, self.session_id, self.start_offset, self.stop_offset
        )

    ### SPECIAL METHODS ###

    def __getitem__(
        self, item: str
    ) -> Union[float, "supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"]:
        assert self.session._active_moments
        offset = self.session._active_moments[-1].offset
        return self._get_at_offset(offset, item)[0] or 0

    def __setitem__(
        self,
        item: str,
        value: Union[float, "supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"],
    ) -> None:
        import supriya.nonrealtime

        assert self.session._active_moments
        offset = self.session._active_moments[-1].offset
        assert isinstance(
            value, (int, float, supriya.nonrealtime.Bus, supriya.nonrealtime.BusGroup)
        )
        self._set_at_offset(offset, item, value)

    ### PRIVATE METHODS ###

    def _add_node(self, node: "Node", add_action: AddActionLike) -> "Node":
        state = self.session._find_state_at(node.start_offset, clone_if_missing=True)
        state.start_nodes.add(node)
        if node not in state.nodes_to_children:
            state.nodes_to_children[node] = None
        state = self.session._find_state_at(node.stop_offset, clone_if_missing=True)
        state.stop_nodes.add(node)
        self.move_node(node, add_action=add_action)
        self.session.nodes.add(node)
        self.session._nodes_by_session_id[node.session_id] = node
        self.session._apply_transitions([node.start_offset, node.stop_offset])
        return node

    def _collect_settings(self, offset: float, *, id_mapping, persistent=False):
        settings: Dict[str, Union[float, str]] = {}
        for key in self._events:
            value, actual_offset = self._get_at_offset(offset, key)
            if not persistent and actual_offset != offset:
                continue
            if id_mapping and value in id_mapping:
                settings[key] = cast(
                    Union["supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"],
                    value,
                ).get_map_symbol(id_mapping[value])
            elif value is not None:
                settings[key] = float(value)
        return settings

    def _fixup_duration(self, new_duration: float) -> None:
        old_duration = self._duration
        if old_duration == new_duration:
            return
        with self.session.at(self.stop_offset, propagate=False) as moment:
            if self in moment.state.stop_nodes:
                moment.state.stop_nodes.remove(self)
            moment.state._sparsify()
        self.session.nodes.remove(self)
        self._duration = new_duration
        self.session.nodes.add(self)
        with self.session.at(self.stop_offset, propagate=False) as moment:
            moment.state.stop_nodes.add(self)

    def _fixup_events(self, new_node: "Node", split_offset: float) -> None:
        left_events: Dict[str, List[Tuple[float, float]]] = {}
        right_events: Dict[str, List[Tuple[float, float]]] = {}
        for name, events in self._events.items():
            for offset, value in events:
                if offset < split_offset:
                    left_events.setdefault(name, []).append((offset, value))
                else:
                    right_events.setdefault(name, []).append((offset, value))
        for name, events in left_events.items():
            if name in right_events and right_events[name][0][0] == split_offset:
                continue
            event = (split_offset, events[-1][-1])
            right_events.setdefault(name, []).insert(0, event)
        self._events = left_events
        new_node._events = right_events

    def _fixup_node_actions(
        self, new_node: "Node", start_offset: "float", stop_offset: "float"
    ) -> None:
        for offset in sorted(self.session.states):
            if offset < start_offset:
                continue
            elif stop_offset < offset:
                break
            transitions = self.session.states[offset].transitions
            if self in transitions:
                transitions[new_node] = transitions.pop(self)
            for node, action in transitions.items():
                if node is new_node:
                    continue
                if action.target is self:
                    action._target = new_node

    def _get_at_offset(
        self, offset: float, item: str
    ) -> Tuple[
        Optional[
            Union[float, "supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"]
        ],
        Optional[float],
    ]:
        """
        Relative to Node start offset.
        """
        events = self._events.get(item)
        if not events:
            return None, None
        index = bisect.bisect_left(events, (offset,))
        if len(events) <= index:
            old_offset, value = events[-1]
        else:
            old_offset, value = events[index]
        if old_offset == offset:
            return value, old_offset
        index -= 1
        if index < 0:
            return None, None
        actual_offset, value = events[index]
        return value, actual_offset

    def _set_at_offset(self, offset, item, value):
        """
        Relative to Synth start offset.
        """
        if offset < self.start_offset or self.stop_offset <= offset:
            return
        events = self._events.setdefault(item, [])
        new_event = (offset, value)
        if not events:
            events.append(new_event)
            return
        index = bisect.bisect_left(events, new_event)
        if len(events) <= index:
            events.append(new_event)
        old_offset, old_value = events[index]
        if old_offset == offset:
            events[index] = (offset, value)
        else:
            events.insert(index, new_event)

    def _split(
        self,
        split_offset: float,
        new_nodes=None,
        split_occupiers: bool = True,
        split_traversers: bool = True,
    ) -> List["Node"]:
        import supriya.nonrealtime

        new_nodes = new_nodes or []
        state = self.session.states[split_offset]
        entering, exiting, occupying, starting, _ = self.inspect_children()
        children = state.nodes_to_children.get(self) or ()
        start_offset, stop_offset = self.start_offset, self.stop_offset
        if start_offset < split_offset < stop_offset:
            old_actions = state.transitions
            new_duration = stop_offset - split_offset
            with supriya.nonrealtime.DoNotPropagate():
                if isinstance(self, supriya.nonrealtime.Synth):
                    new_node = self.add_synth(
                        add_action="ADD_BEFORE",
                        duration=new_duration,
                        synthdef=self.synthdef,
                        **self._synth_kwargs,
                    )
                else:
                    new_node = self.add_group(
                        add_action="ADD_BEFORE", duration=new_duration
                    )
            new_nodes.append(new_node)
            new_actions: Dict["Node", NodeTransition] = collections.OrderedDict()
            for node in new_nodes:
                if node is new_node and self in old_actions:
                    old_actions.pop(node)
                    action = old_actions.pop(self)
                    new_actions[node] = new(action, source=new_node)
                else:
                    new_actions[node] = old_actions.pop(node)
            for child in reversed(children):
                if child in old_actions:
                    old_actions.pop(child)
                action = supriya.nonrealtime.NodeTransition(
                    source=child, target=new_node, action="ADD_TO_TAIL"
                )
                new_actions[child] = action
            new_actions.update(old_actions)
            state._transitions = new_actions
            self._fixup_events(new_node, split_offset)
            self._fixup_duration(split_offset - start_offset)
            self._fixup_node_actions(new_node, split_offset, stop_offset)
            self.session._apply_transitions(
                [new_node.start_offset, new_node.stop_offset]
            )
            result = [self, new_node]
        else:
            return [self]
        for child in children + exiting:
            if (
                (split_occupiers and child in occupying)
                or (split_traversers and child in entering)
                or (split_traversers and child in exiting)
            ):
                child._split(
                    split_offset,
                    new_nodes=new_nodes,
                    split_occupiers=split_occupiers,
                    split_traversers=split_traversers,
                )
        return result

    ### CONSTRUCTORS ###

    @SessionObject.require_offset
    def add_group(
        self,
        add_action: AddActionLike = None,
        duration: Optional[float] = None,
        offset: Optional[float] = None,
    ) -> "supriya.nonrealtime.Group":
        import supriya.nonrealtime

        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError(f"Invalid add action: {add_action}")
        session_id = self.session._get_next_session_id("node")
        node = supriya.nonrealtime.Group(
            self.session, duration=duration, session_id=session_id, start_offset=offset
        )
        self._add_node(node, add_action)
        return node

    @SessionObject.require_offset
    def add_synth(
        self,
        add_action: AddActionLike = None,
        duration: Optional[float] = None,
        synthdef: Optional[SynthDef] = None,
        offset: Optional[float] = None,
        **synth_kwargs,
    ) -> "Synth":
        import supriya.assets.synthdefs
        import supriya.nonrealtime

        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError(f"Invalid add action: {add_action}")
        session_id = self.session._get_next_session_id("node")
        synthdef = synthdef or supriya.assets.synthdefs.default
        node = supriya.nonrealtime.Synth(
            self.session,
            session_id=session_id,
            duration=duration,
            start_offset=offset,
            synthdef=synthdef,
            **synth_kwargs,
        )
        self._add_node(node, add_action)
        return node

    ### MUTATORS ###

    @SessionObject.require_offset
    def move_node(
        self,
        node: "Node",
        add_action: AddActionLike = None,
        offset: Optional[float] = None,
    ) -> "Node":
        import supriya.nonrealtime

        state: State = self.session.active_moments[-1].state
        if state.nodes_to_parents is None:
            state._desparsify()
        if node in state.nodes_to_parents and node in self.get_parentage():
            raise ValueError("Can't add parent as a child.")
        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError("Invalid add action: {add_action}")
        node_action = supriya.nonrealtime.NodeTransition(
            source=node, target=self, action=add_action
        )
        state.transitions[node] = node_action
        self.session._apply_transitions([state.offset, node.stop_offset])
        return node

    def delete(self) -> None:
        start_state = self.session._find_state_at(self.start_offset)
        start_state.start_nodes.remove(self)
        stop_state = self.session._find_state_at(self.stop_offset)
        stop_state.stop_nodes.remove(self)
        start_offset = self.session._find_state_before(
            self.start_offset, with_node_tree=True
        ).offset
        for state_one, state_two in self.session._iterate_state_pairs(
            start_offset, with_node_tree=True
        ):
            state_two._desparsify()
            if self in state_two.nodes_to_children:
                parent = state_two.nodes_to_parents.pop(self)
                inner_children = state_two.nodes_to_children.pop(self) or ()
                outer_children = list(state_two.nodes_to_children[parent])
                index = outer_children.index(self)
                outer_children[index : index + 1] = inner_children
                for child in inner_children:
                    state_two.nodes_to_parents[child] = parent
                state_two.nodes_to_children[parent] = tuple(outer_children) or None
            state_two._transitions = state_two._rebuild_transitions(
                state_one, state_two
            )
            if state_two == self.stop_offset:
                break
        self.session.nodes.remove(self)
        self.session._nodes_by_session_id.pop(self.session_id)
        self.session._apply_transitions([self.start_offset, self.stop_offset])

    @SessionObject.require_offset
    def free(self, offset: float):
        new_duration = offset - self.start_offset
        if new_duration > self.duration:
            raise ValueError("Cannot free after stop offset")
        return self.set_duration(new_duration, clip_children=True)

    def set_duration(self, new_duration: float, clip_children: bool = False) -> "Node":
        import supriya.nonrealtime

        assert new_duration > 0
        if self.duration == new_duration:
            return self
        if new_duration < self.duration:
            split_offset = self.start_offset + new_duration
            if clip_children:
                with self.session.at(split_offset) as moment:
                    old_node, new_node = self.split(
                        split_occupiers=True, split_traversers=True
                    )
                    state = moment.state
                    children = reversed(
                        list(state._iterate_nodes(new_node, state.nodes_to_children))
                    )
                    for child in children:
                        child.delete()
            else:
                with self.session.at(split_offset):
                    old_node, new_node = self.split(
                        split_occupiers=False, split_traversers=False
                    )
                    new_node.delete()
            self.session._find_state_at(new_node.stop_offset)._sparsify()
            return old_node
        else:
            old_stop_offset = self.stop_offset
            new_stop_offset = self.start_offset + new_duration
            with self.session.at(old_stop_offset, propagate=False) as moment:
                parent = self.get_parent()
                moment.state.stop_nodes.remove(self)
                moment.state._sparsify()
            self._fixup_duration(new_duration)
            with self.session.at(old_stop_offset, propagate=False) as moment:
                moment.state._sparsify()
            while parent is not None and parent.stop_offset < new_stop_offset:
                with self.session.at(parent.stop_offset, propagate=False) as moment:
                    action = supriya.nonrealtime.NodeTransition(
                        source=self, target=parent, action="ADD_BEFORE"
                    )
                    moment.state.transitions[self] = action
                    parent = parent.get_parent()
            with self.session.at(new_stop_offset, propagate=False) as moment:
                moment.state.stop_nodes.add(self)
            with self.session.at(self.start_offset, propagate=False) as moment:
                self.session._apply_transitions(moment.state.offset)
            self.session._apply_transitions(
                [self.start_offset, old_stop_offset, new_stop_offset]
            )
            return self

    @SessionObject.require_offset
    def split(
        self,
        split_occupiers: bool = True,
        split_traversers: bool = True,
        offset: Optional[float] = None,
    ) -> List["Node"]:
        if offset is None:
            raise ValueError
        state = self.session.active_moments[-1].state
        self.session._apply_transitions(state.offset)
        shards = self._split(
            offset, split_occupiers=split_occupiers, split_traversers=split_traversers
        )
        stop_offset = shards[-1].stop_offset
        while state is not None and state.offset <= stop_offset:
            self.session._apply_transitions(state.offset)
            state = self.session._find_state_after(state.offset)
        return shards

    ### RELATIONS ###

    @SessionObject.require_offset
    def inspect_children(
        self, offset: Optional[float] = None
    ) -> Tuple[
        Tuple["Node", ...],
        Tuple["Node", ...],
        Tuple["Node", ...],
        Tuple["Node", ...],
        Tuple["Node", ...],
    ]:
        this_state = self.session._find_state_at(offset, clone_if_missing=True)
        prev_state = self.session._find_state_before(this_state.offset, True)
        prev_state._desparsify()
        this_state._desparsify()
        prev_children = prev_state.nodes_to_children.get(self) or ()
        this_children = this_state.nodes_to_children.get(self) or ()
        entering: Set["Node"] = set()
        exiting: Set["Node"] = set()
        occupying: Set["Node"] = set()
        starting: Set["Node"] = set()
        stopping: Set["Node"] = set()
        for node in prev_children:
            if node.stop_offset == offset:
                stopping.add(node)
                continue
            if node in this_children:
                occupying.add(node)
            else:
                exiting.add(node)
        for node in this_children:
            if node.start_offset == offset:
                starting.add(node)
                continue
            if node.stop_offset == offset:
                stopping.add(node)
                continue
            if node in prev_children:
                occupying.add(node)
            else:
                entering.add(node)
        return (
            tuple(sorted(entering, key=lambda x: x.session_id)),
            tuple(sorted(exiting, key=lambda x: x.session_id)),
            tuple(sorted(occupying, key=lambda x: x.session_id)),
            tuple(sorted(starting, key=lambda x: x.session_id)),
            tuple(sorted(stopping, key=lambda x: x.session_id)),
        )

    @SessionObject.require_offset
    def get_parent(self, offset: Optional[float] = None) -> Optional["Node"]:
        state = self.session._find_state_at(offset, clone_if_missing=True)
        if not state.nodes_to_children:
            state = self.session._find_state_before(state.offset, True)
        elif self.stop_offset == state.offset:
            state = self.session._find_state_before(state.offset, True)
        return state.nodes_to_parents.get(self)

    @SessionObject.require_offset
    def get_parentage(self, offset: Optional[float] = None) -> List["Node"]:
        state = self.session._find_state_at(offset, clone_if_missing=True)
        if not state.nodes_to_children:
            state = self.session._find_state_before(state.offset, True)
        node = self
        parentage = [node]
        while state.nodes_to_parents[node] is not None:
            parent = state.nodes_to_parents[node]
            parentage.append(parent)
            node = parent
        return parentage

    ### PUBLIC PROPERTIES ###

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def session_id(self) -> int:
        return self._session_id

    @property
    def start_offset(self) -> float:
        return self._start_offset

    @property
    def stop_offset(self) -> float:
        if self.duration is None:
            return float("inf")
        return self.start_offset + self.duration


class Group(Node):
    """
    A non-realtime group.
    """

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = (
        AddAction.ADD_TO_HEAD,
        AddAction.ADD_TO_TAIL,
        AddAction.ADD_AFTER,
        AddAction.ADD_BEFORE,
    )

    ### SPECIAL METHODS ###

    def __str__(self):
        return "group-{}".format(self.session_id)

    ### PRIVATE METHODS ###

    def _to_request(
        self, action: NodeTransition, id_mapping: Dict[SessionObject, int]
    ) -> GroupNewRequest:
        source_id = id_mapping[action.source]
        target_id = id_mapping[action.target]
        add_action = action.action
        request = GroupNewRequest(
            items=[
                GroupNewRequest.Item(
                    add_action=add_action, node_id=source_id, target_node_id=target_id
                )
            ]
        )
        return request

    def _get_stop_offset(self, offset, event) -> float:
        duration = event.get("duration") or 0
        delta = event.get("delta") or 0
        return offset + max(duration, delta)

    ### PUBLIC METHODS ###

    @SessionObject.require_offset
    def get_children(self, offset: Optional[float] = None) -> List["Node"]:
        state = self.session._find_state_at(offset, clone_if_missing=True)
        if not state.nodes_to_children:
            state = self.session._find_state_before(state.offset, True)
        elif self.stop_offset == state.offset:
            state = self.session._find_state_before(state.offset, True)
        return list(state.nodes_to_children.get(self) or [])


class Synth(Node):
    """
    A non-realtime synth.
    """

    ### CLASS VARIABLES ###

    _valid_add_actions = (AddAction.ADD_BEFORE, AddAction.ADD_AFTER)

    ### INITIALIZER ###

    def __init__(
        self,
        session,
        session_id: int,
        duration: Optional[float] = None,
        synthdef: Optional[SynthDef] = None,
        start_offset: Optional[float] = None,
        **synth_kwargs,
    ) -> None:
        if synthdef is None:
            synthdef = supriya.assets.synthdefs.default
        Node.__init__(
            self, session, session_id, duration=duration, start_offset=start_offset
        )
        self._synthdef = synthdef
        self._synth_kwargs: Dict[str, Any] = synth_kwargs

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        return "synth-{}".format(self.session_id)

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self, offset):
        group = uqbar.graphs.RecordGroup(children=[])
        group.append(
            uqbar.graphs.RecordField("[{}]".format(self.session_id), name="session_id")
        )
        group.append(uqbar.graphs.RecordField(self.synthdef.name))
        for parameter_name in self.synthdef.parameters:
            value = self._get_at_offset(offset, parameter_name)
            field = "{}: {}".format(parameter_name, value)
            group.append(uqbar.graphs.RecordField(label=field))
        return uqbar.graphs.Node(children=[uqbar.graphs.RecordGroup([group])])

    def _collect_settings(
        self, offset: float, *, id_mapping: Dict[Any, float], persistent=False
    ):
        from .buses import Bus, BusGroup

        settings: Dict[str, float] = {}
        parameters = self.synthdef.parameters
        for key in self._events:
            parameter = parameters[key]
            value, actual_offset = self._get_at_offset(offset, key)
            if not persistent and actual_offset != offset:
                continue
            if value is None:
                continue
            if parameter.parameter_rate == ParameterRate.SCALAR or parameter.name in (
                "in_",
                "out",
            ):
                if value in id_mapping:
                    value = id_mapping[value]
                settings[key] = float(value)
            elif isinstance(value, (Bus, BusGroup)) and value in id_mapping:
                settings[key] = cast(
                    Union["supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"],
                    value,
                ).get_map_symbol(id_mapping[value])
            else:
                settings[key] = float(value)
        return settings

    def _get_at_offset(
        self, offset: float, item: str
    ) -> Tuple[
        Optional[
            Union[float, "supriya.nonrealtime.Bus", "supriya.nonrealtime.BusGroup"]
        ],
        Optional[float],
    ]:
        default = self.synthdef.parameters[item].value
        default = self._synth_kwargs.get(item, default)
        value, actual_offset = super()._get_at_offset(offset=offset, item=item)
        return (value or default), actual_offset

    def _to_request(
        self,
        action: NodeTransition,
        id_mapping: Dict[SessionObject, int],
        **synth_kwargs,
    ) -> SynthNewRequest:
        import supriya.nonrealtime

        source_id = id_mapping[action.source]
        target_id = id_mapping[action.target]
        add_action = action.action
        bus_prototype = (supriya.nonrealtime.Bus, supriya.nonrealtime.BusGroup)
        buffer_prototype = (supriya.nonrealtime.Buffer, supriya.nonrealtime.BufferGroup)
        for key, value in synth_kwargs.items():
            if isinstance(value, bus_prototype):
                bus_id = id_mapping[value]
                value = bus_id
                synth_kwargs[key] = value
            elif isinstance(value, buffer_prototype):
                synth_kwargs[key] = id_mapping[value]
        request = SynthNewRequest(
            add_action=add_action,
            node_id=source_id,
            synthdef=self.synthdef.anonymous_name,
            target_node_id=target_id,
            **synth_kwargs,
        )
        return request

    ### PUBLIC PROPERTIES ###

    @property
    def synthdef(self) -> SynthDef:
        return self._synthdef

    @property
    def synth_kwargs(self) -> Dict[str, Any]:
        return self._synth_kwargs.copy()


class RootNode(Group):
    """
    A non-realtime root node.
    """

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = (AddAction.ADD_TO_HEAD, AddAction.ADD_TO_TAIL)

    ### INITIALIZER ###

    def __init__(self, session: "supriya.nonrealtime.Session"):
        Group.__init__(
            self,
            session=session,
            session_id=0,
            duration=float("inf"),
            start_offset=float("-inf"),
        )

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        return "root"

    ### PUBLIC PROPERTIES ###

    @property
    def start_offset(self) -> float:
        return float("-inf")

    @property
    def stop_offset(self) -> float:
        return float("inf")

    @property
    def duration(self) -> float:
        return float("inf")



import collections

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen


class GrainBuf(MultiOutUGen):
    """

    ::

        >>> grain_buf = supriya.ugens.GrainBuf.ar(
        ...     channel_count=2,
        ...     duration=1,
        ...     envelope_buffer_id=-1,
        ...     interpolate=2,
        ...     maximum_overlap=512,
        ...     pan=0,
        ...     position=0,
        ...     rate=1,
        ...     buffer_id=0,
        ...     trigger=0,
        ... )
        >>> grain_buf
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("trigger", 0),
            ("duration", 1),
            ("buffer_id", None),
            ("rate", 1),
            ("position", 0),
            ("interpolate", 2),
            ("pan", 0),
            ("envelope_buffer_id", -1),
            ("maximum_overlap", 512),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class GrainIn(MultiOutUGen):
    """

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> grain_in = supriya.ugens.GrainIn.ar(
        ...     channel_count=2,
        ...     duration=1,
        ...     envelope_buffer_id=-1,
        ...     maximum_overlap=512,
        ...     position=0,
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> grain_in
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("trigger", 0),
            ("duration", 1),
            ("source", None),
            ("position", 0),
            ("envelope_buffer_id", -1),
            ("maximum_overlap", 512),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class PitchShift(UGen):
    """
    A pitch shift unit generator.

    ::

        >>> source = supriya.ugens.SoundIn.ar()
        >>> supriya.ugens.PitchShift.ar(
        ...     source=source,
        ... )
        PitchShift.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("window_size", 0.2),
            ("pitch_ratio", 1.0),
            ("pitch_dispersion", 0.0),
            ("time_dispersion", 0.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Warp1(MultiOutUGen):
    """

    ::

        >>> warp_1 = supriya.ugens.Warp1.ar(
        ...     buffer_id=0,
        ...     channel_count=1,
        ...     envelope_buffer_id=-1,
        ...     frequency_scaling=1,
        ...     interpolation=1,
        ...     overlaps=8,
        ...     pointer=0,
        ...     window_rand_ratio=0,
        ...     window_size=0.2,
        ... )
        >>> warp_1
        Warp1.ar()[0]

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", 0),
            ("pointer", 0),
            ("frequency_scaling", 1),
            ("window_size", 0.2),
            ("envelope_buffer_id", -1),
            ("overlaps", 8),
            ("window_rand_ratio", 0),
            ("interpolation", 1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



import supriya.synthdefs
import supriya.ugens


def _build_default_synthdef():
    r"""
    SynthDef(\default, { arg out=0, freq=440, amp=0.1, pan=0, gate=1;
        var z;
        z = LPF.ar(
            Mix.new(
                VarSaw.ar(
                    freq + [0, Rand(-0.4, 0.0), Rand(0.0, 0.4)],
                    0,
                    0.3,
                    0.3
                    )
                ),
            XLine.kr(Rand(4000,5000), Rand(2500,3200), 1)
            );
        z = z * Linen.kr(gate, 0.01, 0.7, 0.3, 2);
        OffsetOut.ar(out, Pan2.ar(z, pan, amp));
    }, [\ir]).add;
    """

    builder = supriya.synthdefs.SynthDefBuilder(
        amplitude=0.1,
        frequency=440,
        gate=1,
        out=supriya.synthdefs.Parameter(
            parameter_rate=supriya.ParameterRate.SCALAR, value=0
        ),
        pan=0.5,
    )

    with builder:
        low_pass = supriya.ugens.LPF.ar(
            source=supriya.ugens.Mix.new(
                supriya.ugens.VarSaw.ar(
                    frequency=builder["frequency"]
                    + (
                        0,
                        supriya.ugens.Rand.ir(minimum=-0.4, maximum=0.0),
                        supriya.ugens.Rand.ir(minimum=0.0, maximum=0.4),
                    ),
                    width=0.3,
                )
            )
            * 0.3,
            frequency=supriya.ugens.XLine.kr(
                start=supriya.ugens.Rand.ir(minimum=4000, maximum=5000),
                stop=supriya.ugens.Rand.ir(minimum=2500, maximum=3200),
            ),
        )
        linen = supriya.ugens.Linen.kr(
            attack_time=0.01,
            done_action=supriya.DoneAction.FREE_SYNTH,
            gate=builder["gate"],
            release_time=0.3,
            sustain_level=0.7,
        )
        pan = supriya.ugens.Pan2.ar(
            source=low_pass * linen * builder["amplitude"], position=builder["pan"]
        )
        supriya.ugens.OffsetOut.ar(bus=builder["out"], source=pan)
    synthdef = builder.build(name="default")
    return synthdef


default = _build_default_synthdef()

__all__ = ["default"]



from uuid import UUID

from uqbar.objects import get_repr, get_vars, new

from .events import CompositeEvent, Event


class MockUUID:
    def __init__(self, string):
        self.string = string

    def __eq__(self, expr):
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __repr__(self):
        return get_repr(self, multiline=False)


def sanitize_id(id_, cache):
    if id_ not in cache:
        cache[id_] = MockUUID(chr(len(cache) + 65))
    return cache[id_]


def sanitize_event(event, cache):
    if isinstance(event, CompositeEvent):
        return new(event, events=[sanitize_event(x, cache) for x in event.events])
    sanitize_data = {}
    args, _, kwargs = get_vars(event)
    for key, value in args.items():
        if isinstance(value, UUID):
            value = sanitize_id(value, cache)
        sanitize_data[key] = value
    for key, value in sorted(kwargs.items()):
        if isinstance(value, UUID):
            value = sanitize_id(value, cache)
        sanitize_data[key] = value
    return type(event)(**sanitize_data)


def sanitize(exprs):
    cache = {}
    sanitized = []
    for expr in exprs:
        if isinstance(expr, Event):
            sanitized.append(sanitize_event(expr, cache))
        else:
            sanitized.append(expr)
    return sanitized


def run_pattern_test(pattern, expected, is_infinite, stop_at):
    assert pattern.is_infinite == is_infinite
    iterator = iter(pattern)
    actual = []
    ceased = True
    for iteration in range(1000):
        try:
            if stop_at == iteration:
                expr = iterator.send(True)
            else:
                expr = next(iterator)
            actual.append(expr)
        except StopIteration:
            break
    else:
        ceased = False
    if is_infinite:
        assert not ceased
        sanitized_actual = sanitize(actual[: len(expected)])
    else:
        sanitized_actual = sanitize(actual)
    # expected_string = "\n".join(repr(x) for x in expected)
    # actual_string = "\n".join(repr(x) for x in sanitized_actual)
    # for line in unified_diff(expected_string.splitlines(), actual_string.splitlines()):
    #     print(line)
    assert sanitized_actual == expected, sanitized_actual
